#!/etc/opt/openit/perl
#
# $Id: conf_helper.pl,v 1.0.0.0 2020/02/19 03:06:05 cayala Exp $
#
# (C) Copyright 2020 by Open iT Norge AS, All rights reserved.
#
# This software may be used or copied only with the written permission of
# Open Systems Computing AS or in accordance with the terms and conditions
# stipulated in the agreement under which this software has been supplied.
#
# This file is part of Open iT Resources.
#
#HEAD##########################################################################
#
# @NAME:
#   Configuration Helper
# @AUTHOR:
#   Chester Ayala
# @CREATED:
#   Wed Feb 19 20:06:05 MET 2020
# @DESCRIPTION:
#   Utility Perl script subroutines for patching Open iT configuration files.
#
#/HEAD#########################################################################

#MAIN##########################################################################
#
# @HEADING Configuration Helper
#
# This program, which may be included in all patches, will update
# configuration files by calling its subroutines.
#
#/MAIN#########################################################################
my @session = ();
my $file = "";

sub format_conf {
	read_file($file);
	my ($index, $open) = (0, 0);
	
	# Remove newlines, tab whitespaces, and tabbing
	while ($index < @session) {
		@session[$index] =~ s/[\t\n\b]+//g;
		$index++;
	}

	save_file($file);
	read_file($file);

	# Remove space whitespaces
	@session[0] =~ s/(\{[ ]+)/\{/g;
	@session[0] =~ s/(\}[ ]+)/\}/g;
	@session[0] =~ s/(\"[ ]+)/\"/g;

	# Apply newlines
	@session[0] =~ s/\"\{/\"\n\{\n/g;
	@session[0] =~ s/[\"]\}/\}\n/g;
	@session[0] =~ s/\n\}/\n\}\n/g;
	@session[0] =~ s/\}\}/\}\n\}\n/g;
	@session[0] =~ s/\"\n\{\n\}\n/\"\{\}\n/g;
	@session[0] =~ s/\"\{\}\n\}/\"\{\}\n\}\n/g;
	@session[0] =~ s/\n\}\n\}/\n\}\n\}\n/g;
	@session[0] =~ s/\n\n/\n/g;

	save_file($file);
	read_file($file);

	$index = 0;

	# Apply tab whitespaces
	while ($index < @session) {
		if (@session[$index] =~ /\}/ && @session[$index] !~ /\{.*\}/) {
			$open--;
		}

		my $tabs = "";

		for (my $i = 0; $i < $open; $i++) {
			$tabs .= "\t";
		}

		@session[$index] = $tabs . @session[$index];

		if (@session[$index] =~ /\{/ && @session[$index] !~ /\{.*\}/) {
			$open++;
		}

		$index++;
	}

	save_file($file);
	print " -- Applied formatting.\n";
}

sub convert_filterproc
{
	read_file($file);
	my $index = locate_line(0, "filter-process\\\(\\\)\\=\\\"\\\"", \@session);

	if ($index == 0) {
		print " -- filter-process is already converted to filter-list\n";
	} else {
		my $tabs = get_whitespace(@session[$index]);
		@session[$index] =~ s/process/list/;
		@session[$index] .= "$tabs\{\n";
		$index++;
		while (@session[$index] !~ /([\t]+)\}/) {
			my $object = get_object_name(@session[$index]);
			$object = "\t$tabs$object\(\)=\"\"\{\}\n";

			@session[$index] = @session[$index] =~ /([\t]+)\{/ ? "" : $object;

			$index++;
		};

		while (@session[$index] !~ /filter-defs\(\)=\"\"/) {
			@session[$index] = "";
			$index++;
		}

		@session[$index-1] = "\t\}\n";

		save_file($file);
		print " -- Converted filter-process to filter-list\n";
	}
}

sub set_object_value {
	read_file($file);
	my ($target, $value) = (@_);
	my $regex_value = quotemeta($value);
	my ($start_index, $end_index, $mismatch) = check_target_object($target, "normal");

	if ($mismatch == 0) {
		if (@session[$start_index] !~ /\"$regex_value\"/) {
			@session[$start_index] =~ s/\".*\"/\"$value\"/;
			save_file($file);
			print " -- Updated $target\n";
		} else {
			print " -- $target is already updated\n";
		}
	} else {
		print " -- Target: $target not found\n";
	}
}

sub sort_data_generation_config {
	read_file($file);
	my $index = 0;

	while ($index < @session) {
		if (@session[$index] !~ /^#.*$/) {
			if (@session[$index] =~ /^\d+.*$/) {
				for (my $i = $index + 1; @session[$i] =~ /^\d+.*$/; $i++) {
					$session[$index] =~ /^\d+/;
					my $lowest_num = $&;
					@session[$i] =~ /^\d+/;
					my $current_num = $&;
					if ($lowest_num > $current_num) {
						(@session[$index], @session[$i]) = (@session[$i], @session[$index]);
					}
				}
			} elsif (@session[$index] eq "\n" && @session[$index + 1] eq "\n") {
				@session[$index] = "";
			}
		}
		$index++;
	}

	save_file($file);
	print " -- Sorted datatypes\n";
}

# Modes
# - Replace the whole line if $match is left blank
# - Replace $match with $value
sub update_line {
	read_file($file);
	my ($target, $match, $value) = (@_);
	my @parents = split(/\./, $target);
	my ($target_index, $end_index, $regex_value) = (0, 0, quotemeta($value));

	foreach my $element (@parents) {
		$target_index = locate_line($target_index, "^[^#\\n]?.*" . quotemeta($element), \@session);
	}

    my $end_index = $target_index;

    until (@session[$end_index] == "" || $end_index > @session) {
        $end_index++;
    }

	if ($target_index > 0 && $end_index > $target_index) {
		if (@session[$target_index] !~ /$regex_value/) {
			if ($match eq "") {
				@session[$target_index] = "$value\n";
			} else {
				@session[$target_index] =~ s/$match/$value/;
			}
			
			save_file($file);
			print " -- Updated $target\n";
		} else {
			print " -- $target is already updated\n";
		}
	} else {
		print " -- $target doesn't exist\n";
	}
}

# Modes:
# - append
# - prepend
# - insert-after
# - insert-before
sub insert_object {
	read_file($file);
	my ($insert, $target, $mode) = (@_);

	if ($mode =~ /append|prepend|insert-after|insert-before/) {
		my ($object, $last_target) = (get_object_name($insert), "");
		my @parents = split(/\./, $target);
		
		if ($mode =~ /insert-after|insert-before/) {
			$last_target = pop(@parents);
		}
			
		my ($title, $regex_object) = (join(".", @parents, $object), quotemeta($object));
		
		my ($start_index, $end_index, $mismatch) = check_target_object($target, "normal");

		if ($mismatch == 0) {
			my ($object_index, undef, undef) = check_target_object($title, "normal");

			if (@session[$object_index] !~ /$regex_object\(.*\)\=\".*\"/) {
				$insert =~ s/    /\t/g;

				if (@parents == 0) {
					$insert =~ s/\n\t/\n/g;
					if (length($last_target) == 0) {
						$end_index = @session - 1;
						$start_index = 0;
					}
				}

				my $tabs = $mode =~ /insert-after|insert-before/ ? get_whitespace(@session[$end_index]) : get_whitespace(@session[$start_index]);

				if ($mode eq "append") {
					if (@parents == 0) {
						@session[$end_index] .= "$insert\n";
					} else {
						if (@session[$end_index] !~ /\{\}/) {
							$insert =~ s/\n/\n$tabs/g;
							@session[$end_index - 1] .= "\t$tabs$insert\n";
						} else {
							$insert = "\n\{\n\t$insert\n\}";
							$insert =~ s/\n/\n$tabs/g;
							@session[$end_index] =~ s/\{\}/$insert/g;
						}
					}
				} elsif ($mode eq "prepend") {
					if (@parents == 0) {
						@session[$start_index] = "$insert\n" . @session[$start_index];
					} else {
						if (@session[$start_index] !~ /\{\}/) {
							$insert =~ s/\n/\n$tabs/g;
							@session[$start_index + 1] .= "\t$tabs$insert\n";
						} else {
							$insert = "\n\{\n\t$insert\n\}";
							$insert =~ s/\n/\n$tabs/g;
							@session[$start_index] =~ s/\{\}/$insert/g;
						}
					}
				} elsif ($mode eq "insert-after") {
					if (@parents > 0) {
						$tabs =~ s/\t//;
						$insert = "\t$tabs$insert";
						$insert =~ s/\n/\n$tabs/g;
					}
					@session[$end_index] .= "$insert\n";
				} elsif ($mode eq "insert-before") {
					if (@parents > 0) {
						$tabs =~ s/\t//;
						$insert = "\t$tabs$insert";
						$insert =~ s/\n/\n$tabs/g;
					}
					@session[$start_index] = "$insert\n" . @session[$start_index];
				}

				save_file($file);
				print " -- Inserted $title\n";
			} else {
				print " -- $title already exist\n";
			}
		} else {
			print " -- $target is an invalid target\n";
		}
	} else {
		print " -- Invalid mode for inserting object\n";
	}
}

# Modes:
# - append
# - prepend
# - link-after
# - link-before
sub link_object {
	read_file($file);
	my ($insert, $target, $mode) = (@_);

	if ($mode =~ /append|prepend|link-after|link-before/) {
		my ($object, $last_target) = (get_object_name($insert), "");
		my @parents = split(/\./, $target);
		
		if ($mode =~ /link-after|link-before/) {
			$last_target = pop @parents;
		}

		my ($title, $regex_object) = ($target =~ /\*/ ? join(".", @parents, $object) : join(".", @parents, "*.$object"), quotemeta($object));

		my ($start_index, $end_index, $mismatch) = check_target_object($target, "nested");

		if ($mismatch == 0) {
			my ($object_index, undef, undef) = check_target_object($title, "normal");

			if (@session[$object_index] !~ /$regex_object\(.*\)\=\".*\"/) {
				$insert =~ s/    /\t/g;

				if (@parents == 0) {
					$insert =~ s/\n\t/\n/g;
					if (length($last_target) == 0) {
						$end_index = @session - 1;
						$start_index = 0;
					}
				}

				my $tabs = $mode =~ /link-after|link-before/ ? get_whitespace(@session[$end_index]) : get_whitespace(@session[$start_index]);

				if ($mode eq "append") {
					while (@session[$start_index] !~ /\{\}/) {
						$start_index++;
					}

					$tabs = get_whitespace(@session[$start_index]);

					$insert = "\n\{\n\t$insert\n\}";
					$insert =~ s/\n/\n$tabs/g;

					@session[$start_index] =~ s/\{\}/$insert/;
				} else {
					$mode eq "link-before" ? $start_index-- : $start_index++;
					$tabs = get_whitespace(@session[$start_index]);

					$insert =~ s/\{.*\}$/\n\t\{/;
					$insert =~ s/\n.*\}$//;
					$insert =~ s/\n/\n$tabs/g;

					@session[$start_index] .= "$tabs\t$insert\n";
	
					($end_index, $start_index) = get_object_index($start_index + 1, "", "normal");

					$tabs = get_whitespace(@session[$start_index]);

					while ($start_index - 2 <= $end_index) {
						@session[$start_index - 2] = "\t" . @session[$start_index - 2];
						$start_index++;
					}

					@session[$start_index - 3] .= "$tabs\}\n";
				}

				save_file($file);
				print " -- Linked $title\n";
			} else {
				print " -- $title already exist\n";
			}
		} else {
			print " -- $target is an invalid target\n";
		}
	} else {
		print " -- Invalid mode for linking object\n";
	}
}

# Modes:
# - remove
# - unlink
sub remove_object {
	read_file($file);
	my ($target, $mode) = (@_);
	my @parents = split(/\./, $target);

	if ($mode =~ /remove|unlink/) {
		my ($start_index, $end_index, $mismatch) = check_target_object($target, $mode eq "unlink" ? "nested" : "normal");

		if ($mismatch == 0) {
			my $tabs = get_whitespace(@session[$start_index]);

			if ($mode eq "unlink") {
				if (@session[$start_index] =~ /\{\}/) {
					$tabs =~ s/\t//;
					if (@session[$start_index - 1] =~ /^$tabs\{/ && @session[$start_index + 1] =~ /^$tabs\}/) {
						@session[$start_index + 1] =~ s/[\t\b]+//g;
						@session[$start_index - 1] =~ s/[\t\b\n]+//g;
						@session[$start_index - 2] =~ s/[\n]+//g;
					}
				} else {
					@session[$start_index] = "";
					@session[$start_index + 1] = "";
					$start_index += 2;

					while (@session[$start_index] !~ /^$tabs\}/) {
						@session[$start_index] =~ s/\t//;
						$start_index++;
					}
				}
				@session[$start_index] = "";
			} elsif ( $mode eq "remove") {
				my ($open, $inside, $index) = (0, 0, $start_index);

				do {
					if (@session[$index] =~ /\{/ ) {
						$inside = 1;
						$open++;
					}
					if (@session[$index] =~ /\}/ ) {
						$open--;
					}

					@session[$index] = "";

					$index++;
				} until ($open == 0 && $inside == 1);

				$tabs =~ s/\t//;

				if (@session[$start_index - 1] =~ /^$tabs\{/ && @session[$index] =~ /^$tabs\}/) {
					@session[$index] =~ s/[\t\b]+//g;
					@session[$start_index - 1] =~ s/[\t\n\b]+//g;
					@session[$start_index - 2] =~ s/[\n]+//g;
				}
			}

			save_file($file);
			print " -- Removed $target\n";
		} else {
			print " -- $target is already removed\n";
		}
	} else {
		print " -- Invalid mode for removing object\n";
	}
}

sub check_target_object {
	my ($target, $mode) = (@_);
	my @parents = split(/\./, $target);
	my ($start_index, $end_index, $mismatch) = (0, 0, 0);

	if (@session == 0) {
		$mismatch++;
	}

	for (my $i = 0; $i < @parents; $i++) {
		my ($object, $regex_object) = (@parents[$i], quotemeta(@parents[$i]));
		if ($i == 0) {
			if ($object eq "*") {
				if ($i == @parents - 1) {
					$mismatch++;
				} else {
					$i++;
					($end_index, $start_index) = get_object_index($start_index, quotemeta(@parents[$i]), "nested");
				}
			} else {
				$start_index = locate_line($start_index, "^$regex_object\\\(.*\\\)\\=\\\".*\\\"", \@session);
				if (@session[$start_index] =~ /\{\}/) {
					$end_index = $start_index;
				} else {
					$end_index = locate_line($start_index, "^\\\}", \@session);
				}
			}
		} else {
			if ($object eq "*") {
				if ($i == @parents - 1) {
					$mismatch++;
				} else {
					$i++;
					($end_index, $start_index) = get_object_index($start_index, quotemeta(@parents[$i]), "nested");
				}
			} else {
				($end_index, $start_index) = get_object_index($start_index, $regex_object, $mode);
			}
		}

		if ($start_index == 0 && @session[$start_index] !~ /$regex_object\(.*\)\=\".*\"/) {
			$mismatch++;
		}
	}
	
	return ($start_index, $end_index, $mismatch);
}

sub get_object_index {
	my ($start_index, $object, $mode) = (@_);
	my ($index, $open, $inside, $found_index, $found_end_index) = ($start_index, 0, 0, 0, $start_index);
	my $end_index = @session[@session - 1] =~ /\}/ ? @session - 1 : @session - 2;
	do {
		my $current_object = @session[$index];
		if ($current_object !~ /\{\}/) {
			if ($current_object =~ /\{/ ) {
				$inside = 1;
				$open++;
			}
			if ($current_object =~ /\}/ ) {
				$open--;
				if ($open == 1 && $found_end_index == $start_index && $found_index != 0) {
					$found_end_index = $index;
				}
			}
		}

		if ($current_object =~ /$object\(.*\)\=\".*\"/ && ($open == 1 || $mode eq "nested")) {
			$found_index = $index;
			if ($current_object =~ /\{\}/) {
				$found_end_index = $found_index;
			}
		}

		if ($current_object =~ /\{\}/ && ($start_index == $index)) {
			$inside = 1;
		} else {
			$index++;
		}
	} until (($open == 0 && $inside == 1) || $index > $end_index);
	return ($found_end_index, $found_index);
}

sub get_whitespace {
	my ($string) = (@_);
	$string =~ s/[^\t]+//;
	return $string;
}

sub get_object_name {
	my ($object) = (@_);
	my @result = split(/\(/, $object);
	my $name = @result[0];
	$name =~ s/[\t\n\b]+//;
	return $name;
}

sub read_file {
	my ($cfg) = (@_);
	open(IN, "$cfg") || die " -- ERROR: Unable to read '$cfg'!\n";
	my @lines = <IN>;
	close(IN);
	@session = @lines;
	$file = $cfg;
}

# For carets and colons, use the . character to represent them
# For the 3rd argument, always add a \ as prefix to the array
sub locate_line {
	my ($start, $find, $lines_ref) = (@_);
	my ($index) = grep { @$lines_ref[$_] =~ /$find/ } $start ..@$lines_ref;
	return $index;
}

sub save_file {
	my ($cfg) = (@_);
	open(OUT, ">$cfg");
	print OUT @session;
	close(OUT);
	$file = $cfg;
}

1;